"""Classes for defining optimization problem objects for Simulated Annealing (SA)."""

# Author: Genevieve Hayes
# License: BSD 3-clause

from .arith_decay import ArithDecay
from .custom_schedule import CustomSchedule
from .exp_decay import ExpDecay
from .geom_decay import GeomDecay
